import { Promise } from "./promise";
import { delay } from "./timeout";

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Error
function ResolvedQueueError(message) {
  this.name = "resolved";
  if ((message !== undefined) && (typeof message !== "string")) {
    throw new TypeError('You must pass a string.');
  }
  this.message = message || "Default Message";
}
ResolvedQueueError.prototype = new Error();
ResolvedQueueError.prototype.constructor = ResolvedQueueError;

var Queue = function() {
  var promise = this,
    promise_list = [],
    fulfill,
    reject;

  if (!(this instanceof Queue)) {
    return new Queue();
  }

  function canceller() {
    for (var i = 0; i < 2; i++) {
      promise_list[i].cancel();
    }
  }

  Promise.apply(promise, [function (done, fail) {
    fulfill = done;
    reject = fail;
  }, canceller]);

  promise_list.push(delay());
  promise_list.push(promise_list[0].then(function () {
    promise_list.splice(0, 2);
    if (promise_list.length === 0) {
      fulfill();
    }
  }));

  promise.push = function(done, fail) {
    var last_promise = promise_list[promise_list.length - 1],
      next_promise;

    if (this.isRejected || this.isFulfilled) {
      throw new ResolvedQueueError();
    }

    next_promise = last_promise.then(done, fail);
    promise_list.push(next_promise);

    // Handle pop
    promise_list.push(next_promise.then(function (fulfillmentValue) {
      promise_list.splice(0, 2);
      if (promise_list.length === 0) {
        fulfill(fulfillmentValue);
      } else {
        return fulfillmentValue;
      }
    }, function (rejectedReason) {
      if (!(promise.isRejected || promise.isFulfilled)) {
        reject(rejectedReason);
      }
      throw rejectedReason;
    }));

    return this;
  };
};

Queue.prototype = Object.create(Promise.prototype);
Queue.prototype.constructor = Queue;

export { Queue, ResolvedQueueError };
